/* Fakiir
 *
 * (C) Copyright 2005 Denis Rojo <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * "$Id: $"
 *
 */

#include <iostream>
#include <cstdio>
#include <stdarg.h>

#include <fakiir.h>
#include <utils.h>

#define GENERATED_FILE "fakiir-code.cpp"


static FILE *code;

static char code_line[512];

void CODE(char *format, ...) {
  va_list arg;
  va_start(arg,format);
  vsnprintf(code_line, 512, format, arg);
  fputs(code_line, code);
  fputs("\n", code);

  va_end(arg);
}

void methodgen(Method_map *methods) {
  Method_map::iterator m_iter;
  Method *m;

  Argument_map::iterator a_iter;
  Argument *a;

  string argstr;
  int c;
  char conv[64];

  CODE("//////////////////////////////////////////");
  CODE("// Gateway functions for exec() methods");
  CODE("// add below an include to class header");
  for( m_iter = methods->begin(); m_iter != methods->end(); ++m_iter) {
    m = (*m_iter).second; // take the data_type out of the pair
    CODE("void gw_%s (void *obj, Argument_map *args) {", m->name.c_str());
    
    // setup argument string
    argstr = "(\n";

    // fill up arguments
    for(c=0, a_iter = m->args.begin();
	a_iter != m->args.end();
	++a_iter, c++) {

      a = (*a_iter);

      argstr += "\t\t";

      if     (a->type == INT_VALUE)    argstr += "*(int*)";
      else if(a->type == FLOAT_VALUE)  argstr += "*(float*)";
      else if(a->type == STRING_VALUE) argstr += "(char*)";
      
      argstr += "((*args)[";
      sprintf(conv,"%u",c);
      argstr += conv;
      argstr += "]->value)";

      if( a_iter+1 != m->args.end() ) // not the last
	argstr += ",\n";

    }

    argstr += "  )";
    D("gw arg string: %s",argstr.c_str());
    // argstr done
    
    // (( method->parent->name *)obj)-> method->name(
    CODE("\t((%s*)obj)", m->parent->name.c_str());
    CODE("->%s %s;", m->name.c_str(), argstr.c_str());
    CODE("}");
    
  }
  
  CODE("/////////// end of gateway functions");
}

void objectgen(Object *o) {
  Method_map::iterator m_iter;
  Method *m;
  
  Property_map::iterator p_iter;
  Property *p;

  Argument_map::iterator a_iter;
  Argument *a;
  
  // obj = new Object("o->name");
  CODE("obj = new Object(fakiir,\"%s\");",o->name.c_str());

  // loop thru the methods of the object
  for( m_iter = o->methods.begin(); m_iter != o->methods.end(); ++m_iter) {
    m = (*m_iter).second; // take the data_type out of the pair

    // met = new Method(o->methods.*.name);
    CODE("met = new Method(obj, \"%s\");", m->name.c_str());
    
    // assign gateway function to method
    CODE("met->gateway = gw_%s;",m->name.c_str());

    // met->parent = obj;
    //    fputs("met->parent = obj;\n",code);

    // generate the arguments of this method
    for(a_iter = m->args.begin(); a_iter != m->args.end(); ++a_iter) {
      a = (*a_iter);
      
      // arg = new Argument();
      CODE("arg = new Argument();");

      CODE("arg->parse(\"%s %s\");",
	   (a->type==INT_VALUE)?"int ":
	   (a->type==FLOAT_VALUE)?"float ":
	   (a->type==STRING_VALUE)?"char ":
	   "void",        a->name.c_str());



      // met->args.append( arg );
      CODE("met->args.push_back( arg );");
      
      // arg->method = met;
      CODE("arg->method = met;");
      
    }
    
    // obj->methods.append( met );
    CODE("\nobj->methods[ met->name.c_str() ] =  met;");
    

  }

  for(p_iter = o->properties.begin(); p_iter != o->properties.end(); ++p_iter) {
    p = (*p_iter).second; // take out the data_type from the pair

    // prop = new Property();
    CODE("prop = new Property();");
    
    // prop->name = "p->name";
    CODE("prop->name = \"%s\";", p->name.c_str());
    
    // prop->type = "p->type";
    CODE("prop->type = %s;",
	 (p->type==INT_VALUE)?"INT_VALUE":
	 (p->type==FLOAT_VALUE)?"FLOAT_VALUE":
	 (p->type==STRING_VALUE)?"STRING_VALUE":"0");
    
    // obj->properties.append( prop );
    CODE("obj->properties[ prop->name.c_str() ] =  prop;");
    
    // prop->parent = obj;
    CODE("prop->parent = obj;");
    
  }
  
}



////////////////////////////////////////////////////////
// generates the code to reproduce a tree anywhere
void treegen(Object_map *symbols, char *filename) {
  Object_map::iterator o_iter;
  Object *o;

  Object_map::iterator child_iter;
  Object *child;

  code = fopen(GENERATED_FILE,"w"); 
  if(!code) {
    E("can't create file %s", GENERATED_FILE);
    return; }

  CODE("#include <fakiir.h>");  
  CODE("#include <%s>",filename);

  // loop thru objects and use methodgen to generate callbacks
  for( o_iter = symbols->begin(); o_iter != symbols->end(); ++o_iter) {
    o = (*o_iter).second;
    methodgen( &o->methods );
  }

  CODE("Object_map *fakiir_summon(Fakiir *fakiir) {");

  CODE("Object *obj;");
  CODE("Method *met;");
  CODE("Property *prop;");
  CODE("Argument *arg;");

  CODE("Object_map *tree;");
  CODE("tree = new Object_map();");

  for( o_iter = symbols->begin(); o_iter != symbols->end(); ++o_iter) {
    o = (*o_iter).second;
    CODE("///////////////////////////");
    CODE("// instance of %s", o->name.c_str());
    
    objectgen(o);
    
    for(child_iter = o->childs.begin(); child_iter != o->childs.end(); ++child_iter) {
      child = (*child_iter).second; // data_type in the pair

      CODE("{ Object *parent = obj;");
      objectgen( child );
      CODE("parent->childs[ obj->name.c_str() ] = obj;");
      CODE("obj = parent; }");

    }
    
    CODE("(*tree)[ obj->name.c_str() ] = obj;");
    CODE("/////////////////////");
    
    
  }
  CODE("fakiir->tree = tree;");
  CODE("return tree;");
  CODE("}");

  fclose(code);
}

