/* Fakiir
 *
 * (C) Copyright 2005 Denis Rojo <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * "$Id: $"
 *
 */

#include <cstdio>
#include <list>

#include <semantics.h>
#include <fakiir.h>
#include <utils.h>

static Object_map symbols;

static list<Object*> rawlist;

Fakiir fakiir; // our Fakiir for the phase of generation

Object *curr_obj = NULL;
Method *curr_met = NULL;

void open_class(char *name) {
  D("CLASS: %s",name);

  // create the new object
  if(curr_obj) close_class();
  curr_obj = new Object(&fakiir, (string)name);
}

void add_inheritance(char *name) {
  D("CLASS INHERITS FROM %s",name);

  if(!curr_obj) {
    E("CAN'T INHERIT: NO CLASS OPEN");
    return;
  }

  curr_obj->parent = new Object(&fakiir, name);
}

void close_class() {
	if(!curr_obj) {
		E("CAN'T CLOSE CLASS: NONE OPEN");
		return;
	}

	if( curr_met ) close_method();

	if( curr_obj->methods.size() == 0
	 && curr_obj->properties.size() == 0) {
		// discard void class
		delete curr_obj;
		curr_obj = NULL;
		D("CLASS DISCARDED");
	} else {
		// place it in the unsorted list
		rawlist.push_back( curr_obj );
		curr_obj = NULL;
		D("CLASS INCLUDED");
	}
}

void open_method(char *name) {
	if(curr_met) close_method();

	// todo: here should handle return type for method
	// we just strip it off for now
	char *p = name;
	while(*p!=' ') p++;
	p++;

	curr_met = new Method( curr_obj, p );
	D("\tMETHOD: %s",p);
}

void add_argument(char *args) {
	Argument *new_arg;
	if( !curr_obj || !curr_met) return;

	new_arg = new Argument();
	new_arg->parse(args);
	new_arg->method = curr_met;
	curr_met->args.push_back( new_arg );
}

void close_method() {
	if(!curr_met || !curr_obj) return;

	curr_obj->methods[ curr_met->name.c_str() ] =  curr_met;
	curr_met->parent = curr_obj;
	curr_met = NULL;
}

void add_property(char *name) {
	Property *new_prop;

	if( !curr_obj ) return;

	new_prop = new Property();
	new_prop->parse(name);
	new_prop->parent = curr_obj;
	curr_obj->properties[ new_prop->name.c_str() ] =  new_prop; 

	D("\tPROPERTY: %s",name);
}


////////////////////////////////////////////////////////
// parses all the rawlist in a tree based on inheritance
// @TODO: this must be rewritten with support for sorting
// lists in which parents and childs are scattered
Object_map *treesort() {
  list<Object*>::iterator e_iter;
  Object *e;
  Object *s;
  
  if(!rawlist.size()) return NULL;
  
  // put parents in place
  for(e_iter = rawlist.begin(); e_iter != rawlist.end(); ++e_iter) {
    e = (*e_iter);

    if( !e->parent ) {
      // rawlist.remove(e);
      symbols[ e->name.c_str() ] = e;
    }

  }	

  // parse childs left and puts them in place
  for(e_iter = rawlist.begin(); e_iter != rawlist.end(); ++e_iter) {
    Object *found;
    e = (*e_iter);
    // search for parents
    if(!e->parent) continue;
    found = symbols[ e->parent->name.c_str() ];
    if(found) {
      // rawlist.remove(e);
      s->childs[ e->name.c_str() ] = e;
      delete e->parent;
      e->parent = found;
    }
  }

  return &symbols;
}

