/* Fakiir
 *
 * (C) Copyright 2005 Denis Rojo <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <cstdio>
#include <daemons.h>
#include <transfers.h>

///////// XMLRPC

TransXmlRpcMethod::TransXmlRpcMethod(string n, Method *m)
  : XmlRpcServerMethod(n.c_str(), m->parent->fakiir->xmlrpc->srv) {
  fprintf(stderr,"XMLRPC new method %s\n",n.c_str());
  master = m;
}

TransXmlRpcMethod::~TransXmlRpcMethod() { }

void TransXmlRpcMethod::execute(XmlRpcValue& params, XmlRpcValue& result) {
  Argument *a;
  Argument_map::iterator a_iter;
    char tmp[512];
  int c;
  
  if( (unsigned int)params.size() != master->args.size() ) {

    snprintf(tmp,511,"ERROR: call to %s with %u arguments instead of %u expected\n",
	     master->name.c_str(), params.size(), master->args.size());
    fprintf(stderr,tmp);
    result[0] = 0;
    result[1] = tmp;
    return;
  }
  
  for(c=0, a_iter = master->args.begin();
      a_iter != master->args.end();
      ++a_iter, c++) {

    a = *a_iter;
    if      (a->type==INT_VALUE)    *(int*)a->value = (int)params[c];
    else if (a->type==FLOAT_VALUE)  *(float*)a->value = (float)double(params[c]);
    else if (a->type==STRING_VALUE) (char*)a->value = (char*) (std::string(params[c])).c_str();
  }
  
  if( master->exec() ) { // SUCCESS
    result[0] = 1;
    snprintf(tmp,511,"SUCCESS: %s called",
	     master->name.c_str());
    result[1] = std::string(tmp);
  } else {       // FAILURE
    result[0] = 0;
    snprintf(tmp,511,"ERROR: %s returns failure in execution",
	     master->name.c_str());
    result[1] = std::string(tmp);
  }

}

